# -*- coding: utf-8 -*-
from __future__ import absolute_import

"""
oauthlib.oauth1
~~~~~~~~~~~~~~

This module is a wrapper for the most recent implementation of OAuth 1.0 Client
and Server classes.
"""

from .rfc5849 import Client, Server

